@@ -1,4 +1,4 @@
-/* Copyright IBM Corp. 2013, 2019 */
+/* Copyright IBM Corp. 2013, 2020 */
 
 #define _GNU_SOURCE
 
@@ -30,16 +30,6 @@ static void __attribute__((destructor)) qc_destructor(void) {
 		iconv_close(qc_cd);
 }
 
-struct qc_handle *qc_get_cec_handle(struct qc_handle *hdl) {
-	return hdl ? hdl->root : hdl;
-}
-
-struct qc_handle *qc_get_lpar_handle(struct qc_handle *hdl) {
-	for (hdl = hdl->root; hdl != NULL && *(int *)(hdl->layer) != QC_LAYER_TYPE_LPAR; hdl = hdl->next);
-
-	return hdl;
-}
-
 /* Update dbg_level from environment variable */
 static void qc_update_dbg_level(void) {
 	char *s, *end;
@@ -111,7 +101,7 @@ static int qc_debug_file_init(void) {
 				goto out_err;
 			}
 		}
-		qc_debug(NULL, "This is qclib v2.1.0, level 3877f257, date 2020-04-15 22:20:26 +0200\n");
+		qc_debug(NULL, "This is qclib v2.2.1, level 857cc75f, date 2020-10-14 21:58:25 +0200\n");
 	}
 
 	return 0;
@@ -349,21 +339,8 @@ static int qc_hdl_verify(struct qc_handle *hdl, const char *func) {
 
 // De-alloc hdl, leaving out the actual handle
 static void qc_hdl_reinit(struct qc_handle *hdl) {
-	struct qc_handle *ptr = hdl, *arg = hdl;
-
-	while (ptr) {
-		free(ptr->layer);
-		free(ptr->attr_present);
-		free(ptr->src);
-		hdl = ptr->next;
-		if (ptr == arg) {
-			memset(ptr, 0, sizeof(struct qc_handle));
-			ptr->root = ptr;
-		} else
-			free(ptr);
-		ptr = hdl;
-	}
-	qc_hdl_unregister(arg);
+	qc_hdl_prune(hdl);
+	qc_hdl_unregister(hdl);
 }
 
 /** Verifies that either a and (b or c), or none are set. I.e. if only one of the attributes is set, then that's an error */
@@ -377,6 +354,10 @@ static int qc_verify_capped_capacity(struct qc_handle *hdl, enum qc_attr_id a, e
 	if (!val_a && !val_b && !val_c)
 		return 0;
 
+	// Attributes should all be set now
+        if (!val_a || !val_b || !val_c)
+                return 1;
+
 	if ((*val_a && !*val_b && !*val_c) || (!*val_a && (*val_b || *val_c))) {
 		qc_debug(hdl, "Warning: Consistency check (\"capped capacity\") for '%s && (%s || %s)' failed at layer %d (%s/%s): %d && (%d || %d)\n",
 			qc_attr_id_to_char(hdl, a), qc_attr_id_to_char(hdl, b), qc_attr_id_to_char(hdl, c),
@@ -573,6 +554,7 @@ static struct qc_mtype mtypes[] = {
 	{3906, "IBM z14",			"IBM LinuxONE Emperor II"},
 	{8561, "IBM z15",			"IBM LinuxONE III"},
 	{8562, "IBM z15 Model T02",		"IBM LinuxONE III Model LT2"},
+	{3931, "IBM z16",			"IBM LinuxONE Emperor 4"},
 	{0,    NULL,				NULL}
 };
 
@@ -628,8 +610,8 @@ static int qc_post_process_LPAR(struct qc_handle *hdl) {
 }
 
 static int qc_post_process_KVM_host(struct qc_handle *hdl) {
-	struct qc_handle *parent = qc_get_prev_handle(hdl);
-	int *num_conf, rc;
+	struct qc_handle *parent = qc_hdl_get_prev(hdl);
+	int *num_conf, rc, *cps, *ifls;
 
 	qc_debug(hdl, "Fill KVM host layer\n");
 	qc_debug_indent_inc();
@@ -647,13 +629,25 @@ static int qc_post_process_KVM_host(struct qc_handle *hdl) {
 	num_conf = qc_get_attr_value_int(parent, qc_num_core_configured);
 	if (!num_conf)
 		num_conf = qc_get_attr_value_int(parent, qc_num_cpu_configured);
-	rc |= !num_conf ||
-	      qc_set_attr_int(hdl, qc_num_core_total, *num_conf, ATTR_SRC_POSTPROC) ||
-	      qc_copy_attr_value_rename(hdl, qc_num_core_dedicated, parent, qc_num_cpu_dedicated) ||
-	      qc_copy_attr_value_rename(hdl, qc_num_core_shared, parent, qc_num_cpu_shared) ||
-	      qc_copy_attr_value(hdl, parent, qc_num_ifl_total) ||
-	      qc_copy_attr_value(hdl, parent, qc_num_ifl_dedicated) ||
-	      qc_copy_attr_value(hdl, parent, qc_num_ifl_shared);
+	if (!num_conf)
+		rc = 1;
+	if (!rc) {
+		rc |= qc_set_attr_int(hdl, qc_num_core_total, *num_conf, ATTR_SRC_POSTPROC) ||
+		      qc_copy_attr_value_rename(hdl, qc_num_core_dedicated, parent, qc_num_cpu_dedicated) ||
+		      qc_copy_attr_value_rename(hdl, qc_num_core_shared, parent, qc_num_cpu_shared);
+		cps = qc_get_attr_value_int(parent, qc_num_cp_total);
+		ifls = qc_get_attr_value_int(parent, qc_num_cp_total);
+		if (cps && ifls && *cps && *ifls) {
+			// mixed-mode LPARs use CPs only!
+			rc |= qc_set_attr_int(hdl, qc_num_ifl_total, 0, ATTR_SRC_POSTPROC) ||
+			      qc_set_attr_int(hdl, qc_num_ifl_dedicated, 0, ATTR_SRC_POSTPROC) ||
+			      qc_set_attr_int(hdl, qc_num_ifl_shared, 0, ATTR_SRC_POSTPROC);
+		} else {
+			rc |= qc_copy_attr_value(hdl, parent, qc_num_ifl_total) ||
+			      qc_copy_attr_value(hdl, parent, qc_num_ifl_dedicated) ||
+			      qc_copy_attr_value(hdl, parent, qc_num_ifl_shared);
+		}
+	}
 
 	qc_debug_indent_dec();
 
@@ -661,7 +655,7 @@ static int qc_post_process_KVM_host(struct qc_handle *hdl) {
 }
 
 static int qc_post_process_KVM_guest(struct qc_handle *hdl) {
-	struct qc_handle *parent = qc_get_prev_handle(hdl);
+	struct qc_handle *parent = qc_hdl_get_prev(hdl);
 	int rc = 0, *num_conf, *num_CPs, *num_IFLs;
 
 	qc_debug(hdl, "Fill KVM guest layer\n");
@@ -697,9 +691,15 @@ out:
 }
 
 static int qc_post_processing(struct qc_handle *hdl) {
+	struct qc_handle *top_host = NULL;
+	int prune_to_host = 0;
+	char *s, *end;
+
 	qc_debug(hdl, "Post processing: Fill KVM layers\n");
 	qc_debug_indent_inc();
-	for (; hdl; hdl = hdl->next) {
+	for (hdl = qc_hdl_get_root(hdl); hdl; hdl = hdl->next) {
+		if (((int *)(hdl->layer))[1] == QC_LAYER_CAT_HOST)
+			top_host = hdl;
 		switch(*(int *)(hdl->layer)) {
 		case QC_LAYER_TYPE_CEC:
 			if (qc_post_process_CEC(hdl))
@@ -721,6 +721,20 @@ static int qc_post_processing(struct qc_handle *hdl) {
 			break;
 		}
 	}
+
+	if ((s = getenv("QC_PRUNE_TO_HOST")) != NULL) {
+		prune_to_host = strtol(s, &end, 10);
+		if (end != s && prune_to_host != 0) {
+			// Suppress all layers on top of the topmost host above the LPAR layer
+			qc_debug(top_host, "QC_PRUNE_TO_HOST is set\n");
+			if (*(int *)(top_host->layer) != QC_LAYER_TYPE_CEC) {
+				qc_debug_indent_inc();
+				qc_debug(top_host, "Pruning layer %d and above\n", top_host->layer_no);
+				qc_hdl_prune(top_host->next);
+				qc_debug_indent_dec();
+			}
+		}
+	}
 	qc_debug_indent_dec();
 
 	return 0;
@@ -740,8 +754,8 @@ static void *_qc_open(struct qc_handle *hdl, int *rc) {
 	qc_debug(hdl, "_qc_open()\n");
 	qc_debug_indent_inc();
 	*rc = 0;
-	if (qc_new_handle(NULL, &hdl, 0, QC_LAYER_TYPE_CEC) ||
-	    qc_new_handle(hdl, &lparhdl, 1, QC_LAYER_TYPE_LPAR)) {
+	if (qc_hdl_new(NULL, &hdl, 0, QC_LAYER_TYPE_CEC) ||
+	    qc_hdl_new(hdl, &lparhdl, 1, QC_LAYER_TYPE_LPAR)) {
 		*rc = -1;
 		goto out;
 	}
@@ -810,7 +824,7 @@ out:
 	return hdl;
 }
 
-void *qc_open(int *rc) {
+__attribute__ ((visibility ("default"))) void *qc_open(int *rc) {
 	struct qc_handle *hdl = NULL;
 	char *s, *end;
 	int i;
@@ -867,7 +881,7 @@ out:
 	return hdl;
 }
 
-void qc_close(void *hdl) {
+__attribute__ ((visibility ("default"))) void qc_close(void *hdl) {
 	if (qc_hdl_verify(hdl, "qc_close"))
 		return;
 	qc_debug(hdl, "qc_close()\n");
@@ -880,7 +894,7 @@ void qc_close(void *hdl) {
 	qc_debug_indent_dec();
 }
 
-int qc_get_num_layers(void *cfg, int *rc) {
+__attribute__ ((visibility ("default"))) int qc_get_num_layers(void *cfg, int *rc) {
 	struct qc_handle *hdl = cfg;
 
 	if (qc_hdl_verify(hdl, "qc_get_num_layers")) {
@@ -914,7 +928,7 @@ static int qc_is_attr_id_valid(enum qc_attr_id id) {
 	return id <= qc_secure;
 }
 
-int qc_get_attribute_string(void *cfg, enum qc_attr_id id, int layer, const char **value) {
+__attribute__ ((visibility ("default"))) int qc_get_attribute_string(void *cfg, enum qc_attr_id id, int layer, const char **value) {
 	struct qc_handle *hdl;
 	int rc;
 
@@ -950,7 +964,7 @@ out:
 	return rc;
 }
 
-int qc_get_attribute_int(void *cfg, enum qc_attr_id id, int layer, int *value) {
+__attribute__ ((visibility ("default"))) int qc_get_attribute_int(void *cfg, enum qc_attr_id id, int layer, int *value) {
 	struct qc_handle *hdl;
 	void *ptr = NULL;
 	int rc;
@@ -992,7 +1006,7 @@ out:
 }
 
 
-int qc_get_attribute_float(void *cfg, enum qc_attr_id id, int layer, float *value) {
+__attribute__ ((visibility ("default"))) int qc_get_attribute_float(void *cfg, enum qc_attr_id id, int layer, float *value) {
 	struct qc_handle *hdl;
 	void *ptr = NULL;
 	int rc;
@@ -1032,3 +1046,33 @@ out:
 
 	return rc;
 }
+
+static void qc_start_object(int *jindent, int layer) {
+	printf("%*s\"Layer %d\": {\n", *jindent, "", layer);
+	*jindent += 2;
+}
+static void qc_end_object(int *jindent, int final) {
+	*jindent -= 2;
+	printf("%*s}%s\n", *jindent, "", (final ? "" : ","));
+}
+
+__attribute__ ((visibility ("default"))) void qc_export_json(void *cfg) {
+	struct qc_handle *hdl = (struct qc_handle *)cfg;
+	int jindent = 0;	// indent for json output
+	int i;
+
+	if (!hdl)
+		return;
+
+	printf("{\n");
+	jindent += 2;
+	for (hdl = hdl->root, i = 0; hdl != NULL; hdl = hdl->next, i++) {
+		qc_start_object(&jindent, i);
+		qc_print_attrs_json(hdl, jindent);
+		qc_end_object(&jindent, hdl->next == NULL);
+	}
+
+	printf("}\n");
+
+	return;
+}
